%{
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "agatparse.h"
#include "agatcheck.h"

%}


%union {
     FADef         *pDef;
     FAConstruct   *pConstruct;
     FAReg         *pReg;
     FAName        *pNames;
     FAClause      *pClause;
     FAPattern     *pPattern;
     Byte  State;
     FAAction      *pAction;
     FAExpNode     *pExpNod;
     char          *pStr;
     Value         *pVal;
}


%token   PAR_OP  PAR_CL  BRA_OP  BRA_CL  SBR_OP  SBR_CL 
%token   COMMA  SEMCO  ARROW  ASSIGN  AFFECT  PIPE
%token   SHARP  CIRCU  STAR  UNDER  MINUS  PLUS  DIV 
%token   MAP  LET  IF  THEN  ELSE  ENDIF  NULL_ACTION  INCLUDE
%token   GT  LT  GTOEQ  LTOEQ  EQUAL  NEQUA

%token   <pStr>   IDENT  STRING
%token   <pVal>   NUMBER  TOK_TRUE  TOK_FALSE

%type   <pDef>       defconst defmacro defstream defpostproc
%type   <pConstruct> construct map
%type   <pReg>       registers regdecls regdecl
%type   <pNames>     names name 
%type   <pClause>    clauses clause
%type   <pPattern>   patterns pattern
%type   <State>      state
%type   <pAction>    actions action assignation ifthenelse
%type   <pExpNod>    expression sexpression factors
%type   <pExpNod>    funcall fact factor params


%%


all: definitions 
   | {fprintf(stderr,"warning: file %s seems to be empty!\n",curFile);}
   ;


definitions: definition definitions
           | definition
           ;

include: INCLUDE files 
       ;

files: file COMMA files
     | file
     ;

file: STRING   {addInclude($1);}
    ;

definition: defconst  SEMCO {addDef($1);}
          | defmacro  SEMCO {addDef($1);}
          | defstream   SEMCO {addDef($1);}
          | defpostproc SEMCO {addDef($1);}
          | include   SEMCO
          | error     SEMCO
          | SEMCO
          ;

defconst: LET IDENT ASSIGN expression  {$$=bDef(DT_CONST,$2,NULL,NULL,$4);}
        | IDENT ASSIGN expression  {$$=bDef(DT_CONST,$1,NULL,NULL,$3);}
        ;

defmacro: LET IDENT PAR_OP names PAR_CL ASSIGN construct  {$$=bDef(DT_MACRO,$2,$4,$7,NULL);}
        | IDENT PAR_OP names PAR_CL ASSIGN construct  {$$=bDef(DT_MACRO,$1,$3,$6,NULL);}
        ;

defstream: LET IDENT ASSIGN construct    {$$=bDef(DT_STREAM,$2,NULL,$4,NULL);}
         | IDENT ASSIGN construct    {$$=bDef(DT_STREAM,$1,NULL,$3,NULL);}
         ;

defpostproc: IDENT PAR_OP names PAR_CL   {$$=bDef(DT_POSTPROC,$1,$3,NULL,NULL);}
           ;

construct: map        {$$=$1;}
         ;

map: MAP registers PAR_OP clauses PAR_CL names                                       {$$=bConstruct(CT_MAP, $2,$4,$6);}
   | MAP PAR_OP clauses PAR_CL names                                                 {$$=bConstruct(CT_MAP, NULL,$3,$5);}
   ;

registers: BRA_OP regdecls BRA_CL {$$=$2;}
         ;

regdecls: regdecl COMMA regdecls   {$1->next=$3;$$=$1;}
        | regdecl                  {$$=$1;}
        ;

regdecl: IDENT                     {$$=bReg($1,NULL);}
       | IDENT AFFECT expression   {$$=bReg($1,$3);}
       ;

names: name COMMA names   {$1->next=$3;$$=$1;}
     | name               {$$=$1;}
     ;

name: IDENT    {$$=bName($1);}
    ;

clauses: clause PIPE clauses   {$1->next=$3;$$=$1;}
       | clause                {$$=$1;}
       ;

clause: patterns ARROW actions   {$$=bClause($1,$3);}
      ;

patterns: pattern COMMA patterns   {$1->next=$3;$$=$1;}
        | pattern                  {$$=$1;}
        ;

pattern: IDENT state   {$$=bPattern($1,$2);}
       | state IDENT   {$$=bPattern($2,$1);}
       | IDENT         {$$=bPattern($1,ST_NO_STATE);}
       | state         {$$=bPattern(NULL,$1);}
       ;

state: SHARP   {$$=ST_NO_VALUE;}
     | CIRCU   {$$=ST_FIRST_NEW_VALUE;}
     | UNDER   {$$=ST_NEW_VALUE;}
     | STAR    {$$=ST_NO_NEW_VALUE;}
     | PLUS    {$$=ST_NO_NEW_VALUE_BUT_ONE;}
     ;

actions: action COMMA actions   {$1->next=$3; $$=$1;}
       | action                 {$$=$1;}
       ;

action: expression    {$$=bAction(ACT_EXPR,NULL,$1,NULL,NULL);}
      | ifthenelse    {$$=$1;}
      | assignation   {$$=$1;}
      | NULL_ACTION   {$$=bAction(ACT_NULL,NULL,NULL,NULL,NULL);}
      ;

ifthenelse: IF expression THEN actions ENDIF                                                {$$=bAction(ACT_IFTE,NULL,$2,$4,NULL);}
          | IF expression THEN actions ELSE actions ENDIF                                   {$$=bAction(ACT_IFTE,NULL,$2,$4,$6);}
          ;

assignation: IDENT AFFECT expression   {$$=bAction(ACT_ASGN,$1,$3,NULL,NULL);}
           ;

expression: sexpression                    {$$=$1;}                   
          | sexpression EQUAL sexpression  {$$=benFuncBin(OPEQUAL,$1,$3);}
          | sexpression NEQUA sexpression  {$$=benFuncBin(OPNEQUA,$1,$3);}  
          | sexpression GT sexpression     {$$=benFuncBin(OPGT,$1,$3);}  
          | sexpression LT sexpression     {$$=benFuncBin(OPLT,$1,$3);}  
          | sexpression GTOEQ sexpression  {$$=benFuncBin(OPGTOEQ,$1,$3);}  
          | sexpression LTOEQ sexpression  {$$=benFuncBin(OPLTOEQ,$1,$3);}  
          ;

sexpression: factors                      {$$=$1;}                    
           | sexpression PLUS factors     {$$=benFuncBin(OPPLUS,$1,$3);} 
           | sexpression MINUS factors    {$$=benFuncBin(OPMINUS,$1,$3);} 
           ;

factors: factor                  {$$=$1;}    
       | factors STAR factor     {$$=benFuncBin(OPMULT,$1,$3);}  
       | factors DIV factor      {$$=benFuncBin(OPDIV,$1,$3);}
       ;
                          

factor: fact            {$$=$1;}
      | PLUS fact       {$$=$2;}
      | MINUS fact      {$$=benFuncUn(OPMINUS_U,$2);}
      ;

fact: NUMBER                    {$$=benValue($1);} 
    | IDENT                     {$$=benVar($1);}
    | TOK_TRUE                      {$$=benValue($1);}
    | TOK_FALSE                     {$$=benValue($1);}
    | funcall                   {$$=$1;}
    | PAR_OP expression PAR_CL  {$$=$2;}
    ;

funcall: IDENT PAR_OP PAR_CL          {$$=benFuncCall($1,NULL); }
       | IDENT PAR_OP params PAR_CL   {$$=benFuncCall($1,$3); }
       ;
       
params: params COMMA expression {$$=benArgList($3,$1); }
      | expression              {$$=benArgList($1,NULL); }
      ;


%%

#include "agatlex.c"

yyerror(s)
    char           *s;
{
    char            tmp[MAX_CHAR_TMP];

    sprintf(tmp, "%s%s", ER_YACC_SYNTAX, yytext);
    parseError(tmp);
}


int yywrap()
{
  int             fd, i,a;
  IncDesc        *pi;
  
  for (i = 0; i < htNbElts(htInclude); i++) {
    pi = htLookNth(htInclude, i);
    if (pi->done == False) {
      fd = UOpenPath(lIncludePath,pi->name, O_RDONLY);
      if (fd < 0){
	checkError(pi->file,pi->lgn,UOpenError(errno,pi->name));
	pi->done = True;
	continue; /* for the next include */
      }
      close(0);
      dup(fd);
      close(fd);
      yyin=fdopen(0,"r");
      curLgn = 1;
      curFile = pi->name;
      pi->done = True;
      return 0;
    }
  }
  return 1;
}

